"""Define constants for the Ambient PWS component."""
DOMAIN = "ambient_station"

ATTR_LAST_DATA = "last_data"
ATTR_MONITORED_CONDITIONS = "monitored_conditions"

CONF_APP_KEY = "app_key"

DATA_CLIENT = "data_client"

TOPIC_UPDATE = "ambient_station_data_update_{0}"

TYPE_BINARY_SENSOR = "binary_sensor"
TYPE_SENSOR = "sensor"
