"""Define constants for the Spotify integration."""

DOMAIN = "spotify"

CONF_CLIENT_ID = "client_id"
CONF_CLIENT_SECRET = "client_secret"

DATA_SPOTIFY_CLIENT = "spotify_client"
DATA_SPOTIFY_ME = "spotify_me"
DATA_SPOTIFY_SESSION = "spotify_session"
