"""Constants for the Epson projector component."""
DOMAIN = "epson"
SERVICE_SELECT_CMODE = "select_cmode"

ATTR_CMODE = "cmode"

DATA_EPSON = "epson"
DEFAULT_NAME = "EPSON Projector"

SUPPORT_CMODE = 33001
